# add the channel conda-forge to the environment my_env in anaconda navigator (run as timperle)
# then run conda install pywin32==225 in the terminal (specifically to work with Python 3.7/3.8)

def send_email(my_subject, my_body, **kwargs):
    
    import win32com.client
    import configparser
    from config.config_defs import CONFIG_PATH

    my_attachment = None

    for key, value in kwargs.items():
        if key == 'my_attachment':
            my_attachment = value

    config = configparser.ConfigParser()		
    config.read(CONFIG_PATH)
    my_email = config['email']
    my_to = my_email["to"]
    
    outlook = win32com.client.Dispatch('outlook.application')
    
    mail = outlook.CreateItem(0)
    
    mail.To = my_to
    # mail.CC = 'somebody@company.com'
    mail.Subject = my_subject
    mail.Body = my_body
    # mail.HTMLBody = '<h3>This is HTML Body</h3>'
    
    if my_attachment != None:
        mail.Attachments.Add(my_attachment)
    
    mail.Send()

def write_to_log(log_path, my_text, **kwargs):
    
    import datetime
    # from gen_mods.misc_fns import send_email
    
    email_update = False
    email_subject = None

    for key, value in kwargs.items():
        if key == 'email_update':
            email_update = value
        if key == 'email_subject':
            email_subject = value
            
    # create output text
    output_text = datetime.datetime.now().strftime("%d/%m/%Y: %H:%M:%S") + ': ' + my_text
    
    # print to screen
    print(output_text)
    
    # write to log
    log_file = open(log_path, 'a')
    log_file.write(output_text + '\n')
    log_file.close()
    
    if email_update == True:
        if email_subject != None:
            send_email(email_subject, my_text, my_attachment = log_path)
        else:
            send_email('automated email update', my_text, my_attachment = log_path)
            
def exec_fn(my_fn, my_log):

    import sys
    
    my_stem = '* function ' + my_fn.__name__ + ' in ' + my_fn.__module__
    write_to_log(my_log, my_stem + ' started *')

    # my_fn()
    
    try:
        my_fn()
    except Exception as e:        
        err_msg = '*** code execution failed -- error below ***'  + '\r' + str(e)    
        if my_log != None and len(my_log) > 0:
            write_to_log(my_log, err_msg)
        sys.exit('code execution failed -- check log file')    
    else:    
        if my_log != None and len(my_log) > 0:
            write_to_log(my_log, my_stem + ' succeeded *')
    
def create_empty_log_file(log_path):
    
    log_file = open(log_path, 'a')
    log_file.seek(0)
    log_file.truncate()
    log_file.close()
    
def run_subprocess(file_path, log_path, **kwargs):
    
    import subprocess
    import sys
    from gen_mods.misc_fns import write_to_log

    failure_email = False

    for key, value in kwargs.items():
        if key == 'failure_email':
            failure_email = value
            
    if log_path != None:
        write_to_log(log_path, 'subprocess started: ' + file_path)
    
    r = subprocess.run([file_path], shell=True)

    if r.returncode != 0:
        err_msg = 'subprocess failed: ' + file_path
        if failure_email == True:
            if log_path != None:
                write_to_log(log_path, err_msg, email_update=True, email_subject='code execution failed -- check log file')
        else:
            if log_path != None:
                write_to_log(log_path, err_msg)               
        sys.exit('code execution failed -- check log file')
    else:
        if log_path != None:
            write_to_log(log_path, 'subprocess succeeded: ' + file_path)
        
def create_empty_text_file(my_path):
    
    log_file = open(my_path, 'a')
    log_file.seek(0)
    log_file.truncate()
    log_file.close()
    
def append_to_text_file(my_path, my_text):
     
    # print to screen
    print(my_text)
    
    # write to log
    log_file = open(my_path, 'a')
    log_file.write(my_text + '\n')
    log_file.close()
    
def write_to_new_text_file(my_path, my_text):
    
    create_empty_text_file(my_path)
    append_to_text_file(my_path, my_text)
    
def current_method_name():
    
    import inspect
    
    # [0] is this method's frame and [1] is its parent's
    return inspect.stack()[1].function

def path_is_in_syspath(path):
    
    import os
    import sys
    
    path = os.path.normcase(path)
    return any(os.path.normcase(sp) == path for sp in sys.path)

def remove_from_syspath(path):

    import os
    import sys
    
    # print('old sys path:',sys.path)
    
    while path_is_in_syspath(path) == True:
        sys.path.remove(path)
    
    # print('new sys path:',sys.path)

def append_to_syspath(path):

    import os
    import sys
    
    print('old sys path:',sys.path)
    
    remove_from_syspath(path)    
    sys.path.append(path)
    
    print('new sys path:',sys.path)
    
def remove_from_python_path(path_to_remove):

    import os
    import sys
         
    python_path = os.environ['PYTHONPATH']
    print('old python path:', python_path)
    
    path_list = python_path.split(os.pathsep)
    path_list = [x for x in path_list if x]
    
    if path_to_remove in path_list:
        os.environ['PYTHONPATH'] = ';'.join([x for x in path_list if x != path_to_remove])    
    print('new python path:', os.environ['PYTHONPATH'])    

def append_to_python_path(path_addendum):

    # path_addendum = r'C:\my_local_folder\my_apps\my_python\xx' 
    # ^^^ leave commented out
    
    import os
    import sys
    
    python_path = os.environ['PYTHONPATH']
    print('old python path:', python_path)
    
    path_list = python_path.split(os.pathsep)    
    path_list = [x for x in path_list if x]
        
    if not path_addendum in path_list:
        os.environ['PYTHONPATH'] = python_path + ';' + path_addendum    
    print('new python path:', os.environ['PYTHONPATH'])


from handyfuncs.automation import (
    write_to_log
    , exec_fn
    , create_empty_log_file
    , run_subprocess
    , create_empty_text_file
    , append_to_text_file
    , write_to_new_text_file
    , current_method_name
    )

from handyfuncs.manipulation import (
    number_of_words_in_string
    , all_same_character
    , get_initials
    , literal_return
    , explode_frame_list_column
    , convert_dict_to_list
    , merge_dictionaries
    )


from setuptools import setup 

with open("README.md", 'r') as f:
    long_description = f.read()

setup( 
    name='handyfuncs', 
    version='0.0.1', 
    description='Handy functions for manipulating data and automating processes',
    author='Porte Verte', 
    author_email='porte_verte@outlook.com', 
    url='https://github.com/porteverte/handyfuncs',
    packages=['handyfuncs'],
    package_dir={'':'src'},
    python_requires=">=3.8",
)


def number_of_words_in_string(my_string):
    
    # make sure we are dealing with a string (in case a float is passed)
    my_string = str(my_string)
    
    # determine the number of words in the string by splitting it into into a list and getting its length    
    return len(my_string.strip().split())

def all_same_character(s, **kwargs):
    
    for k, v in kwargs.items():
        if k == 'c':
            c = v
    
    n = len(s)
    for i in range(1, n):
        if s[i] != c:
            return False
        if s[i] != s[0]:
            return False
 
    return True

def get_initials(s):
    # try:
    return ''.join([x[0] for x in str(s).split()])
    # except AttributeError:
    #     return s

def literal_return(val):
    
    from ast import literal_eval
    
    try:
        return literal_eval(val)
    except (ValueError, SyntaxError) as e:
        return val

def explode_frame_list_column(my_frame, my_column):
        
    # remove blank values -- and make a copy to keep the original frame in tact
    my_frame = my_frame[my_frame[my_column].str.len() > 0].copy()
    
    # evaluate the column -- because the lists are stored as text
    my_frame[my_column] = my_frame[my_column].apply(lambda x: literal_return(str(x)))
    
    # now do the actual explosion
    my_frame = my_frame.explode(my_column)
    
    return my_frame

def convert_dict_to_list(my_dict):
# turns the company suffixes dicitonary into a list to be used in company standardisation    
    my_list = []
    for md in my_dict:
        for my_item in my_dict[md]:
            my_list.append((my_item, md))
    # sort the list by the descending length of the first tuple value
    # this is so that replacments are made to the longest strings first in case one shorter term is inside another longer one 
    # e.g. kg is also at the end of gmbh & co kg so if it found a variant of kg first it would not replace all of the term
    my_list.sort(key=lambda x: len(x[0]), reverse=True)
    return my_list

def merge_dictionaries(d1, d2):
    
    # make copies of dictionaries to avoid messing them up
    d1_copy = d1.copy()
    d2_copy = d2.copy()
    
    # create a list comprising all keys from both dictionaries (convert list to set then back to list to dedupe it)
    my_keys = list(set([*d1_copy] + [*d2_copy]))
    
    # if a key is missing from one of the dictionaries then add it (then both will have exactly the same keys)
    for m in my_keys:
        if not m in d1_copy:
            d1_copy.update({m: None})
        if not m in d2_copy:
            d2_copy.update({m: None})
    
    # create blank dictionary
    d = {}
    
    # add each item in the keys list to the dictionary as a key with None as the value 
    for i in my_keys:
        d[i] = None
    
    # merge the values for the keys in each dictionary and add them to the new one
    for k in d.keys():
        # merge the lists of values (will produce a list of lists)
        merged_lists = [d1_copy[k],d2_copy[k]]
        # remove NoneType from lists
        clean_list = [x for x in merged_lists if x is not None]
        # flatten the list of lists into one plain list
        flat_list = [x for xs in clean_list for x in xs]
        # dedupe the list
        deduped_list = list(set(flat_list))
        # add the list to the dictionary
        d[k] = deduped_list
    
    return d